#include <list>
#include <queue>
#include <math.h>
#include <rgb.hpp>
#include <medianCut.hpp>

// Median cut
// CTOR
Quantizer::MedianCut::MedianCut(const unsigned int& iColorCount,
                                const unsigned int& iBitCount) :
   colorCount(iColorCount),
   bitCount  (iBitCount),
   tree      (),
   palette   ()
{}
// DTOR
Quantizer::MedianCut::~MedianCut() {
   tree.Clear();
   if(!palette.empty())
	palette.clear();
}

// Add a color
void Quantizer::MedianCut::Add(const unsigned char& iR,
             		       const unsigned char& iG,
		               const unsigned char& iB) {
    // Add color to octree
    tree.Add(Color::RGB(iR, iG, iB));
}

// Perform quantization
void Quantizer::MedianCut::Realize() {
   std::priority_queue<Quantizer::MedianCut::Box> boxQueue;

   std::vector<Quantizer::MedianCut::Octree::ColorBookEntry>::iterator begin(tree.FirstColor());
   std::vector<Quantizer::MedianCut::Octree::ColorBookEntry>::iterator end (tree.LastColor());
   std::vector<Quantizer::MedianCut::Octree::ColorBookEntry>::iterator median;

   // Top-bottom median cut tree creation
   Quantizer::MedianCut::Box box(begin, end, tree.ColorCount());

   boxQueue.push(box);
   while(boxQueue.size() < colorCount) {
	// Get the box with the "longest line"
	Quantizer::MedianCut::Box current = boxQueue.top();
	boxQueue.pop();
	// Compute median
	size_t medianIndex = ((current.Count() + 1)/ 2);
	begin  = current.First();
	end    = current.Last ();
        median = current.First() + medianIndex;

	// Partial sort along the longest line axis
	switch(current.Axis()) {
	    case 0:
                std::nth_element(begin, median, end, 
				 Quantizer::MedianCut::Octree::ColorBookEntry::WeakComparator<0>());
                break;
            case 1:
                std::nth_element(begin, median, end,
                                 Quantizer::MedianCut::Octree::ColorBookEntry::WeakComparator<1>());
                break;
            case 2:
		 std::nth_element(begin, median, end,
                                 Quantizer::MedianCut::Octree::ColorBookEntry::WeakComparator<2>());
                 break;
	}
	// Split
	Quantizer::MedianCut::Box left (begin,  median, medianIndex);
	Quantizer::MedianCut::Box right(median, end,    current.Count() - medianIndex);
	boxQueue.push(left);	
	boxQueue.push(right);
   }

   // Build colorbook from the queue
   off_t colorIndex = 0;
   while(!boxQueue.empty()) {
	box = boxQueue.top();
	std::vector<Quantizer::MedianCut::Octree::ColorBookEntry>::iterator it;	
	long int dummy[3];
	dummy[0] = dummy[1] = dummy[2] = 0;
	for(it=box.First(); it!=box.Last(); ++it) {
		dummy[0] += it->color.r;
		dummy[1] += it->color.g;
		dummy[2] += it->color.b;
        	it->node->color = colorIndex;
        }
	dummy[0] /= box.Count();
	dummy[1] /= box.Count();
	dummy[2] /= box.Count();

	++colorIndex;
	palette.push_back(Color::RGB(dummy[0], dummy[1], dummy[2]));
	boxQueue.pop();
   }    
}

// Data accessors
unsigned int Quantizer::MedianCut::ColorCount() const {
	return colorCount;
}
unsigned int Quantizer::MedianCut::BitCount() const {
	return bitCount;
}

// Compute palette
unsigned char* Quantizer::MedianCut::Palette(unsigned int& iColorCount) {
	unsigned char* clut = new unsigned char[colorCount * 3];
        off_t idx;

	iColorCount = colorCount;

        // Create palette from octree
	std::vector<Color::RGB>::iterator it;
	idx = 0;
	for(it=palette.begin(); it!=palette.end(); ++it) {
	    clut[idx++] = it->r;
	    clut[idx++] = it->g;
	    clut[idx++] = it->b;
        }

	return clut;
}

// Get palette index
unsigned int Quantizer::MedianCut::Index(const unsigned char& iR,
		                         const unsigned char& iG,
                                         const unsigned char& iB) {
        // Traverse octree to find corresponding index
	return tree.Index(Color::RGB(iR,iG,iB));
}


// Color Axis aligned box
// CTOR
Quantizer::MedianCut::Box::Box(std::vector<Quantizer::MedianCut::Octree::ColorBookEntry>::iterator& iFirst,
                               std::vector<Quantizer::MedianCut::Octree::ColorBookEntry>::iterator& iLast,
			       const size_t&                      iCount) :
  first (iFirst),
  last  (iLast),
  count (iCount),
#ifndef USE_VARIANCE
  length(0),
#endif
  index (0)
{
  unsigned int j;
#ifdef USE_VARIANCE
  long int mean[3];
  mean[0] = mean[1] = mean[2] = 0;
#endif
  // Compute the bounding box enclosing the point cloud
  min = max = first->color;
  std::vector<Quantizer::MedianCut::Octree::ColorBookEntry>::iterator it;
  for(it=first+1; it!=last; ++it) {
    for(j=0; j<3; ++j) {
      if(min.c[j] > it->color.c[j]) {
	min.c[j] = it->color.c[j];
      }
      if(max.c[j] < it->color.c[j]) {
	max.c[j] = it->color.c[j];
      }
#ifdef USE_VARIANCE
      mean[j] += it->color.c[j];
#endif
    }
  }

#ifdef USE_VARIANCE
  mean[0] /= count;
  mean[1] /= count;
  mean[2] /= count;

  variance[0] = variance[1] = variance[2] = 0.0;
  for(it=first; it!=last; ++it) {
    for(j=0; j<3; ++j) {
      variance[j] += pow(static_cast<double>(it->color.c[j] - mean[j]), 2);
    }
  }

  variance[0] /= count;
  variance[1] /= count;
  variance[2] /= count;
  for(j=1; j<3; ++j) {
    if(variance[j] > variance[index])
      index = j;
  }
#else
  // Compute longest side index and length
  length = max.c[0] - min.c[0];
  for(j=1; j<3; ++j) {
    uint8_t diff = max.c[j] - min.c[j];
    if(diff > length) {
      length = diff;
      index  = j;
    }
  }
#endif
}

Quantizer::MedianCut::Box::Box(const Quantizer::MedianCut::Box& box) :
  first (box.first),
  last  (box.last),
  count (box.count),
  min   (box.min),
  max   (box.max),
#ifndef USE_VARIANCE
  length(box.length),
#endif
  index (box.index)
{
#ifdef USE_VARIANCE
  variance[0] = box.variance[0];
  variance[1] = box.variance[1];
  variance[2] = box.variance[2];
#endif
}
// DTOR
Quantizer::MedianCut::Box::~Box() {}
// =
Quantizer::MedianCut::Box& Quantizer::MedianCut::Box::operator=(const Quantizer::MedianCut::Box& box) {
  first  = box.first;
  last   = box.last;
  count  = box.count;
  min    = box.min;
  max    = box.max;
  index  = box.index;
#ifndef USE_VARIANCE
  length = box.length;
#else
  variance[0] = box.variance[0];
  variance[1] = box.variance[1];
  variance[2] = box.variance[2];
#endif
  return *this;
}

// Get data
std::vector<Quantizer::MedianCut::Octree::ColorBookEntry>::iterator
Quantizer::MedianCut::Box::First() const {
  return first;
}
std::vector<Quantizer::MedianCut::Octree::ColorBookEntry>::iterator 
Quantizer::MedianCut::Box::Last() const {
  return last;
}
size_t Quantizer::MedianCut::Box::Count() const {
  return count;
}
Color::RGB Quantizer::MedianCut::Box::Min() const {
  return min;
}
Color::RGB Quantizer::MedianCut::Box::Max() const {
  return max;
}
uint8_t Quantizer::MedianCut::Box::Axis() const {
  return index;
}
#ifdef USE_VARIANCE
double Quantizer::MedianCut::Box::Variance(const off_t& i) const {
  return variance[i];
}
#else
uint8_t Quantizer::MedianCut::Box::Length() const {
  return length;
}
#endif
// Octree
// CTOR
Quantizer::MedianCut::Octree::Octree() :
  nodeList(),
  colorBook()
{
  root = dynamic_cast<Quantizer::MedianCut::Octree::Node*>(new Quantizer::MedianCut::Octree::Internal());
  nodeList.push_back(root);
}
// DTOR
Quantizer::MedianCut::Octree::~Octree() {
  Clear();
}
// Delete tree
void Quantizer::MedianCut::Octree::Clear() {
  // Delete node list
  std::list<Quantizer::MedianCut::Octree::Node*>::iterator it;
  for(it=nodeList.begin(); it!=nodeList.end(); ++it) {
    delete (*it);
  }
  nodeList.clear();
  // Delete color book
  colorBook.clear();
}
// Add color to tree
void Quantizer::MedianCut::Octree::Add(const Color::RGB& color) {
  // Color book creation
  Internal* current;
  Leaf*     last;
  uint8_t   bit, index;
  // Traverse tree (and create internal nodes)
  current = dynamic_cast<Quantizer::MedianCut::Octree::Internal*>(root);
  for(bit=0; bit<7; ++bit) {
    index = (((color.r >> (7-bit)) & 1) << 2) |
            (((color.g >> (7-bit)) & 1) << 1) |
            (((color.b >> (7-bit)) & 1));

    if(!current->child[index]) {
      current->child[index] = dynamic_cast<Quantizer::MedianCut::Octree::Node*>(new Quantizer::MedianCut::Octree::Internal());
      nodeList.push_back(current->child[index]); 
    }
    current = dynamic_cast<Quantizer::MedianCut::Octree::Internal*>(current->child[index]);
  }

  // Leaf set up
  index = ((color.r & 1) << 2) |
          ((color.g & 1) << 1) |
          ((color.b & 1));
  if(!current->child[index]) {
    current->child[index] = dynamic_cast<Quantizer::MedianCut::Octree::Node*>(new Leaf());
    nodeList.push_back(current->child[index]);
  }

  last = dynamic_cast<Quantizer::MedianCut::Octree::Leaf*>(current->child[index]);
  if(!last->color) {
    // We found a new color to add to the color book
    last->Add(colorBook.size());
    colorBook.push_back(Quantizer::MedianCut::Octree::ColorBookEntry(color,last));
  }
  ++last->occurence;
}
// Get color index
unsigned int Quantizer::MedianCut::Octree::Index (const Color::RGB& color) {
  unsigned int i, index;
  Quantizer::MedianCut::Octree::Node* tmp = root;
  i = 0;
  do {
    if(!tmp->IsLeaf()) {
      index = (((color.r >> (7-i)) & 1) << 2) |
              (((color.g >> (7-i)) & 1) << 1) |
              (((color.b >> (7-i)) & 1));

       Quantizer::MedianCut::Octree::Internal* current =
           dynamic_cast<Quantizer::MedianCut::Octree::Internal*>(tmp);
       tmp = current->child[index];
     } else {
       return dynamic_cast<Quantizer::MedianCut::Octree::Leaf*>(tmp)->color;
     }
     ++i;
    } while(tmp);
    return 0;
}
// Returns an iterator to the begin of the color array
std::vector<Quantizer::MedianCut::Octree::ColorBookEntry>::iterator
Quantizer::MedianCut::Octree::FirstColor() {
  return colorBook.begin();
}
// Returns an iterator to the end of the color array
std::vector<Quantizer::MedianCut::Octree::ColorBookEntry>::iterator
Quantizer::MedianCut::Octree::LastColor() {
  return colorBook.end();
}
// Color count
size_t Quantizer::MedianCut::Octree::ColorCount() const {
  return colorBook.size(); 
}

// Octree::Node
// CTOR
Quantizer::MedianCut::Octree::Node::Node() {}
// DTOR
Quantizer::MedianCut::Octree::Node::~Node() {}

// Octree:Internal
// CTOR
Quantizer::MedianCut::Octree::Internal::Internal() {
  child[0] = child[1] = child[2] = child[3] =
  child[4] = child[5] = child[6] = child[7] = 0;
}
// DTOR
Quantizer::MedianCut::Octree::Internal::~Internal() {}
// An internal node at least has one child 
bool Quantizer::MedianCut::Octree::Internal::IsLeaf() {
  return false;
}

// Octree::Leaf
// CTOR
Quantizer::MedianCut::Octree::Leaf::Leaf() :
  color    (0),
  occurence(0)
{}
// CTOR
Quantizer::MedianCut::Octree::Leaf::~Leaf() {}
// A leaf is a leaf :)
bool Quantizer::MedianCut::Octree::Leaf::IsLeaf() {
  return true;
}
void Quantizer::MedianCut::Octree::Leaf::Add(const off_t& colorIndex) {
  color = colorIndex;
}

// -- Color book entry
// CTOR
Quantizer::MedianCut::Octree::ColorBookEntry::ColorBookEntry() : 
  color(),
  node (0) 
{}
Quantizer::MedianCut::Octree::ColorBookEntry::ColorBookEntry(const Color::RGB& iColor, 
							     Quantizer::MedianCut::Octree::Leaf* iLeaf) :
  color(iColor),
  node (iLeaf)
{}
Quantizer::MedianCut::Octree::ColorBookEntry::ColorBookEntry(const Quantizer::MedianCut::Octree::ColorBookEntry& iCEntry) :
  color(iCEntry.color),
  node (iCEntry.node)
{}
// DTOR
Quantizer::MedianCut::Octree::ColorBookEntry::~ColorBookEntry() {}
// =
Quantizer::MedianCut::Octree::ColorBookEntry& 
Quantizer::MedianCut::Octree::ColorBookEntry::operator= (const Quantizer::MedianCut::Octree::ColorBookEntry& iCEntry) {
  color = iCEntry.color;
  node  = iCEntry.node;
  return *this;
}

