#ifndef _MEDIAN_CUT_HPP_
#define _MEDIAN_CUT_HPP_

#include <list>
#include <vector>
#include <rgb.hpp>

namespace Quantizer {
  // Median Cut color quantization
  class MedianCut {
  public:
    // CTOR
    MedianCut(const unsigned int&, const unsigned int& = 8);
    // DTOR
    ~MedianCut();

    // Add a color
    void Add(const unsigned char&,
             const unsigned char&,
             const unsigned char&);
    // Perform quantization
    void Realize();
    // Data accessors
    unsigned int ColorCount() const;
    unsigned int BitCount() const;
    // Compute palette
    unsigned char* Palette(unsigned int&);
    // Get palette index
    unsigned int Index(const unsigned char&,
                       const unsigned char&,
                       const unsigned char&);

  private:
    // Color octree (used to get color histogram and unique colors)
    class Octree {
    public:
      // CTOR
      Octree();
      // DTOR
      ~Octree();
      // Delete tree
      void Clear();
      // Add color to tree
      void Add(const Color::RGB&);
      // Color count
      size_t ColorCount() const;
      // Get color index
      unsigned int Index(const Color::RGB&);

    private:
      // Octree node
      struct Node {
	// CTOR
	Node();
	// DTOR
	virtual ~Node();
	// Is this a leaf?
	virtual bool IsLeaf() = 0;
      };
      // -- Internal node
      struct Internal : public Node {
	// CTOR
	Internal();
	// DTOR
	~Internal();
	// Well no
	bool IsLeaf();
	// Children
	Node* child[8];
      };
      // -- Leaf
      struct Leaf : public Node {
	// CTOR
	Leaf();
	// DTOR
	~Leaf();
	// It guess the answer'ld be yes
	bool IsLeaf();
	// Add color
	void Add(const off_t&);
	// Color associated to the node
	off_t color;
	// Number of occurence in the image
	unsigned long int occurence;
      };

      // -- Color book entry
      struct ColorBookEntry {
        // CTOR
        ColorBookEntry();
        ColorBookEntry(const Color::RGB&, Leaf*);
        ColorBookEntry(const ColorBookEntry&);
        // DTOR
        ~ColorBookEntry();
        // =
        ColorBookEntry& operator= (const ColorBookEntry&);
        // Strict week ordering
        template<int index>
        struct WeakComparator {
           bool operator() ( const ColorBookEntry& iC0,
                             const ColorBookEntry& iC1) {
               return iC0.color.c[index] < iC1.color.c[index];
           }
        };
        // Color
        Color::RGB color;
        // Associated leaf
        Leaf*      node;
      };

      public:
        // Returns an iterator to the begin of the color array
        std::vector<ColorBookEntry>::iterator FirstColor();
          // Returns an iterator to the end of the color array
        std::vector<ColorBookEntry>::iterator LastColor();
      private:
      // Root node
      Node* root;
      // Internal node list
      std::list<Node*> nodeList;
      // Color vector
      std::vector<ColorBookEntry> colorBook; 
    };

    // Axis aligned box containing points in color space
    class Box {
    public:
      // CTOR
      Box(std::vector<Octree::ColorBookEntry>::iterator&,
          std::vector<Octree::ColorBookEntry>::iterator&,
          const size_t&);
      Box(const Box&);
      // DTOR
      ~Box();
      // =
      Box& operator=(const Box&);

      // Comparators (LessThan comparable)
#ifdef USE_VARIANCE
      friend bool operator< (const Box& iB0, const Box& iB1)
        { return (iB0.variance[iB0.index] < iB1.variance[iB1.index]); }
      friend bool operator> (const Box& iB0, const Box& iB1)
        { return (iB0.variance[iB0.index] > iB1.variance[iB1.index]); }
#else
      friend bool operator< (const Box& iB0, const Box& iB1)
	{ return (iB0.length < iB1.length); }
      friend bool operator> (const Box& iB0, const Box& iB1)
	{ return (iB0.length > iB1.length); }
#endif
      // Access data
      std::vector<Octree::ColorBookEntry>::iterator First() const;
      std::vector<Octree::ColorBookEntry>::iterator Last () const;
      size_t      Count () const;
      Color::RGB  Min   () const;
      Color::RGB  Max   () const;
      uint8_t     Axis  () const;
#ifdef USE_VARIANCE
      double      Variance(const off_t&) const;
#else
      uint8_t     Length() const;
#endif
    private:
      // Points contained in the box
      std::vector<Octree::ColorBookEntry>::iterator first;
      std::vector<Octree::ColorBookEntry>::iterator last;
      // Element count
      size_t      count;
      // Box coordinates
      Color::RGB  min;
      Color::RGB  max;
#ifdef USE_VARIANCE
      // Variance along each coordiante
      double      variance[3];
#else
      // Length of the longest side
      uint8_t     length;
#endif
      // Axis of the longest side
      uint8_t     index;
    };

    // Desired color count
    unsigned int colorCount;
    // Desired bit count
    unsigned int bitCount;
    // Octree (for color book creation)
    Octree tree;
    // Final palette
    std::vector<Color::RGB> palette;
  };
};

#endif // _MEDIAN_CUT_HPP_
